@@ -1,7 +1,7 @@
 /*
  *
- *  mooproxy - a buffering proxy for moo-connections
- *  Copyright (C) 2001-2007 Marcel L. Moreaux <marcelm@luon.net>
+ *  mooproxy - a buffering proxy for MOO connections
+ *  Copyright (C) 2001-2011 Marcel L. Moreaux <marcelm@qvdr.net>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -30,6 +30,7 @@
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <stdlib.h>
+#include <limits.h>
 
 #include "global.h"
 #include "network.h"
@@ -47,7 +48,7 @@
 #define NET_BACKLOG 4
 
 /* Authentication messages */
-#define NET_AUTHSTRING "Welcome, please authenticate."
+#define NET_AUTHSTRING "Welcome, this is mooproxy. Please authenticate."
 #define NET_AUTHFAIL "Authentication failed, goodbye."
 
 
@@ -62,6 +63,9 @@ static void verify_authentication( World *, int );
 static void promote_auth_connection( World *, int );
 static void handle_client_fd( World * );
 static void handle_server_fd( World * );
+static void privileged_add( World *, char * );
+static void privileged_del( World *, char * );
+static int is_privileged( World *, char * );
 
 
 
@@ -69,16 +73,27 @@ extern void wait_for_network( World *wld )
 {
 	struct timeval tv;
 	fd_set rset, wset;
-	int high, i;
+	int high, i, r;
 
-	/* If there is a correctly authenticated authconn, promote that
-	 * connection and be done. */
+	/* Check the auth connections for ones that need to be promoted to
+	 * client, and for ones that need verification. */
 	for( i = 0; i < wld->auth_connections; i++ )
-		if( wld->auth_correct[i] )
+	{
+		switch( wld->auth_status[i] )
 		{
+			case AUTH_ST_CORRECT:
 			promote_auth_connection( wld, i );
 			return;
+
+			case AUTH_ST_VERIFY:
+			if( wld->auth_tokenbucket > 0 || wld->auth_ispriv[i] )
+			{
+				verify_authentication( wld, i );
+				return;
+			}
+			break;
 		}
+	}
 
 	/* If there are unprocessed lines in the *_toqueue queues, return.
 	 * The main loop will move these lines to their respective transmit
@@ -92,10 +107,16 @@ extern void wait_for_network( World *wld )
 	tv.tv_sec = 1;
 	tv.tv_usec = 0;
 
-	if( select( high + 1, &rset, &wset, NULL, &tv ) < 0 )
-		/* None of select()s errors should occur. */
-		panic( PANIC_SELECT, errno, 0 );
-
+	r = select( high + 1, &rset, &wset, NULL, &tv );
+	if( r < 0 )
+	{
+		/* EINTR is ok, other errors are not. */
+		if( errno == EINTR )
+			return;
+		else
+			panic( PANIC_SELECT, errno, 0 );
+	}
+	
 	/* Check all of the monitored FD's, and handle them appropriately. */
 
 	if( wld->server_resolver_fd != -1 &&
@@ -282,7 +303,7 @@ extern void world_bind_port( World *wld, long port )
 			goto fail_this_af;
 
 		/* Try IPV6_V6ONLY, so IPv4 won't get mapped onto IPv6 */
-		#if defined( PF_INET6 ) && defined( IPV6_V6ONLY )
+		#if defined( AF_INET6 ) && defined( IPV6_V6ONLY )
 		if( ai->ai_family == AF_INET6 && setsockopt( fd, IPPROTO_IPV6,
 				IPV6_V6ONLY, &yes, sizeof( yes ) ) < 0 )
 			goto fail_this_af;
@@ -433,6 +454,7 @@ static void handle_connecting_fd( World *wld )
 {
 	int so_err, fd = wld->server_connecting_fd;
 	socklen_t optlen = sizeof( so_err );
+	Line *line;
 
 	/* Get the SO_ERROR option from the socket. This option holds the
 	 * error code for the connect() in the case of connection failure.
@@ -471,8 +493,10 @@ static void handle_connecting_fd( World *wld )
 	/* Flag and announce connectedness. */
 	wld->server_status = ST_CONNECTED;
 
-	world_msg_client( wld, "      Succes." );
-	world_msg_client( wld, "Connected to world %s.", wld->name );
+	world_msg_client( wld, "      Success." );
+	line = world_msg_client( wld, "Now connected to world %s (%s:%s).",
+			wld->name, wld->server_host, wld->server_port );
+	line->flags = LINE_CHECKPOINT;
 
 	/* Inform the MCP layer that we just connected to the server. */
 	world_mcp_server_connect( wld );
@@ -546,6 +570,11 @@ extern void world_disconnect_server( World *wld )
 
 extern void world_disconnect_client( World *wld )
 {
+	/* We don't want to mess up the next client with undesired ansi
+	 * stuff, so we always disable ace on disconnect. */
+	if( wld->ace_enabled )
+		world_disable_ace( wld );
+
 	if( wld->client_fd != -1 )
 		close( wld->client_fd );
 
@@ -605,21 +634,42 @@ static void handle_listen_fd( World *wld, int which )
 	/* FIXME: Should this be logged somewhere? */
 	/* printf( "Accepted connection from %s.\n", hostbuf ); */
 
-	/* If all auth slots are full, kick oldest */
+	/* If all auth slots are full, we need to kick one out. */
 	if( wld->auth_connections == NET_MAXAUTHCONN )
-		remove_auth_connection( wld, 0, 0 );
+	{
+		int privcount = 0, remove = 0;
+
+		for( i = 0; i < NET_MAXAUTHCONN; i++ )
+			privcount += wld->auth_ispriv[i];
+
+		/* If there are many privconns, just select the oldest conn. */
+		if( privcount > NET_AUTH_PRIVRES )
+			remove = 0;
+		/* If there are few privconns, select the oldest non-priv. */
+		else
+			for( i = 0 ; i < NET_MAXAUTHCONN; i++ )
+				if( !wld->auth_ispriv[i] )
+				{
+					remove = i;
+					break;
+				}
+
+		/* Remove the selected connection. */
+		remove_auth_connection( wld, remove, 0 );
+	}
 
 	/* Use the next available slot */
 	i = wld->auth_connections++;
 
-	/* Initialize carefully. auth_correct = 0 is especially important. */
+	/* Initialize carefully. auth_status is especially important. */
 	wld->auth_fd[i] = newfd;
-	wld->auth_correct[i] = 0;
+	wld->auth_status[i] = AUTH_ST_WAITNET;
 	wld->auth_read[i] = 0;
 	wld->auth_address[i] = xstrdup( hostbuf );
+	wld->auth_ispriv[i] = is_privileged( wld, wld->auth_address[i] );
 
 	/* Zero the buffer, just to be sure.
-	 * Now an erroneous compare won't check against garbage (or, worse,
+	 * Now an erroneous compare won't check against garbage (or worse,
 	 * a correct authentication string) still in the buffer. */
 	memset( wld->auth_buf[i], 0, NET_MAXAUTHLEN );
 
@@ -637,8 +687,8 @@ static void handle_auth_fd( World *wld, int wa )
 {
 	int i, n;
 
-	/* If this connection has already been authenticated, ignore it. */
-	if( wld->auth_correct[wa] )
+	/* Only proceed if this connection is actually waiting for more data */
+	if( wld->auth_status[wa] != AUTH_ST_WAITNET )
 		return;
 
 	/* Read into the buffer */
@@ -668,9 +718,10 @@ static void handle_auth_fd( World *wld, int wa )
 		return;
 	}
 
-	/* We encountered a newline or read too many characters */
+	/* We encountered a newline or read too many characters, verify the
+	 * authentication attempt */
 	wld->auth_read[wa] += n;
-	verify_authentication( wld, wa );
+	wld->auth_status[wa] = AUTH_ST_VERIFY;
 }
 
 
@@ -687,6 +738,18 @@ static void remove_auth_connection( World *wld, int wa, int donack )
 		write( wld->auth_fd[wa], NET_AUTHFAIL "\r\n",
 				sizeof( NET_AUTHFAIL "\r\n" ) - 1);
 
+	/* Connections that have been actively denied are counted towards
+	 * failed login attempts, so we record some information. */
+	if( donack )
+	{
+		if( wld->client_login_failures < LONG_MAX )
+			wld->client_login_failures++;
+		wld->client_last_failtime = current_time();
+		free( wld->client_last_failaddr );
+		wld->client_last_failaddr = wld->auth_address[wa];
+		wld->auth_address[wa] = NULL;
+	}
+
 	/* Free allocated resources */
 	free( wld->auth_address[wa] );
 	if( wld->auth_fd[wa] != -1 )
@@ -702,7 +765,8 @@ static void remove_auth_connection( World *wld, int wa, int donack )
 		wld->auth_read[i] = wld->auth_read[i + 1];
 		wld->auth_buf[i] = wld->auth_buf[i + 1];
 		wld->auth_address[i] = wld->auth_address[i + 1];
-		wld->auth_correct[i] = wld->auth_correct[i + 1];
+		wld->auth_status[i] = wld->auth_status[i + 1];
+		wld->auth_ispriv[i] = wld->auth_ispriv[i + 1];
 	}
 
 	wld->auth_connections--;
@@ -711,7 +775,8 @@ static void remove_auth_connection( World *wld, int wa, int donack )
 	wld->auth_fd[wld->auth_connections] = -1;
 	wld->auth_buf[wld->auth_connections] = buf;
 	wld->auth_address[wld->auth_connections] = NULL;
-	wld->auth_correct[wld->auth_connections] = 0;
+	wld->auth_status[wld->auth_connections] = AUTH_ST_WAITNET;
+	wld->auth_ispriv[wld->auth_connections] = 0;
 }
 
 
@@ -724,14 +789,19 @@ static void verify_authentication( World *wld, int wa )
 {
 	char *buffer = wld->auth_buf[wa];
 	int alen, maxlen, buflen = wld->auth_read[wa];
+	Line *line;
 
 	/* If the authentication string is nonexistent, reject everything */
-	if( wld->auth_md5hash == NULL )
+	if( wld->auth_hash == NULL )
 	{
 		remove_auth_connection( wld, wa, 1 );
 		return;
 	}
 
+	/* Each authentication attempt uses up one token. */
+	if( wld->auth_tokenbucket > 0 )
+		wld->auth_tokenbucket--;
+
 	/* Determine the maximum number of characters to scan */
 	maxlen = buflen + 2;
 	if( maxlen > NET_MAXAUTHLEN )
@@ -745,6 +815,8 @@ static void verify_authentication( World *wld, int wa )
 	/* If we didn't find a newline, trash it. */
 	if( buffer[alen] != '\n' )
 	{
+		if( wld->auth_ispriv[wa] )
+			privileged_del( wld, wld->auth_address[wa] );
 		remove_auth_connection( wld, wa, 1 );
 		return;
 	}
@@ -759,6 +831,8 @@ static void verify_authentication( World *wld, int wa )
 	/* Now, check if the string before the newline is correct. */
 	if( !world_match_authentication( wld, buffer ) )
 	{
+		if( wld->auth_ispriv[wa] )
+			privileged_del( wld, wld->auth_address[wa] );
 		remove_auth_connection( wld, wa, 1 );
 		return;
 	}
@@ -777,15 +851,28 @@ static void verify_authentication( World *wld, int wa )
 
 	/* Tell the client (if any) that the connection is taken over,
 	 * and flag the client connection for disconnection. */
-	if( wld->client_fd != -1 )
+	if( wld->client_status == ST_CONNECTED )
 	{
-		world_msg_client( wld, "Connection taken over by %s.",
+		/* We don't want to mess up the next client with undesired
+		 * stuff, so we disable ace now. */
+		if( wld->ace_enabled )
+			world_disable_ace( wld );
+
+		line = world_msg_client( wld, "Connection taken over by %s.",
 				wld->auth_address[wa] );
+		line->flags &= ~LINE_DONTLOG;
 		wld->flags |= WLD_CLIENTQUIT;
 	}
+	else
+	{
+		/* Record this connection for the log. */
+		line = world_msg_client( wld, "Client connected from %s.",
+				wld->auth_address[wa] );
+		line->flags = LINE_LOGONLY;
+	}
 
 	/* Finally, flag this connection as correctly authenticated. */
-	wld->auth_correct[wa] = 1;
+	wld->auth_status[wa] = AUTH_ST_CORRECT;
 }
 
 
@@ -796,7 +883,7 @@ static void verify_authentication( World *wld, int wa )
 static void promote_auth_connection( World *wld, int wa )
 {
 	/* If the world is not correctly authenticated, ignore it. */
-	if( !wld->auth_correct[wa] )
+	if( wld->auth_status[wa] != AUTH_ST_CORRECT )
 		return;
 
 	/* The client shouln't be connected.
@@ -814,6 +901,10 @@ static void promote_auth_connection( World *wld, int wa )
 	wld->auth_fd[wa] = -1;
 	wld->client_address = wld->auth_address[wa];
 	wld->auth_address[wa] = NULL;
+	wld->client_connected_since = current_time();
+
+	/* Add this address to the list of privileged addresses. */
+	privileged_add( wld, wld->client_address );
 
 	/* In order to copy anything left in the authbuffer to the rxbuffer,
 	 * the rxbuffer must be large enough. */
@@ -829,31 +920,43 @@ static void promote_auth_connection( World *wld, int wa )
 	/* Clean up stuff left of the auth connection */
 	remove_auth_connection( wld, wa, 0 );
 
-	/* Confirm the connection to the authenticating client */
-	world_msg_client( wld, "Authentication successful, welcome." );
+	/* Introduce ourselves, and offer help. We're polite! */
+	world_msg_client( wld, "This is mooproxy %s. Get help with: %shelp.",
+			VERSIONSTR, wld->commandstring );
 
 	/* If a client was previously connected, say when and from where. */
 	if( wld->client_prev_address != NULL )
-	{
 		world_msg_client( wld, "Last connected at %s, from %s.",
-				time_string( wld->client_last_connected,
-				FULL_TIME ), wld->client_prev_address );
-	}
-	else
-	{
-		/* Otherwise, this is the first connect. Display mooproxy
-		 * version and "use /help..." message. */
+				time_fullstr( wld->client_last_connected ),
+				wld->client_prev_address );
 
-		/* FIXME: actually enable this.
-		 * Probably after /help has been improved. */
-		/* world_msg_client( wld, "This is mooproxy " VERSIONSTR
-				". Use %shelp to get online help.",
-				wld->commandstring ); */
+	/* If there were unsuccessful login attempts since the last login,
+	 * report on that. */
+	if( wld->client_login_failures > 0 )
+	{
+		world_msg_client( wld, "" );
+		world_msg_client( wld, "%i failed login attempt%s since %s.",
+				wld->client_login_failures,
+				wld->client_login_failures > 1 ? "s" : "",
+				wld->client_prev_address != NULL ?
+				"your last login" : "mooproxy started" );
+		world_msg_client( wld, "Last failed attempt at %s, from %s.",
+				time_fullstr( wld->client_last_failtime ),
+				wld->client_last_failaddr );
+		world_msg_client( wld, "" );
 	}
 
 	/* Show context history and pass buffered text */
 	world_recall_and_pass( wld );
 
+	/* If there were a lot of failed login attempts, warn the user. */
+	if( wld->client_login_failures >= NET_TOOMANY_LOGIN_FAILURES )
+		world_msg_client( wld, "POSSIBLE BREAK-IN ATTEMPT! Many login"
+			" failures, scroll up for more info." );
+
+	/* Reset the unsuccessful login counter. */
+	wld->client_login_failures = 0; 
+
 	/* Inform the MCP layer that a client just connected. */
 	if( wld->server_fd != -1 )
 		world_mcp_client_connect( wld );
@@ -867,6 +970,7 @@ static void handle_client_fd( World *wld )
 {
 	int n, offset = wld->client_rxfull;
 	char *buffer = wld->client_rxbuffer;
+	Line *line;
 
 	/* Read into the buffer */
 	n = read( wld->client_fd, buffer + offset, NET_BBUFFER_LEN - offset );
@@ -875,10 +979,19 @@ static void handle_client_fd( World *wld )
 	if( n == -1 && ( errno == EINTR || errno == EAGAIN ) )
 		return;
 
+	/* The connection died, record a message. */
 	if( n < 1 )
 	{
-		/* Ok, the connection dropped dead */
+		if( n < 0 )
+			line = world_msg_client( wld, "Connection to client "
+					"lost (%s).", strerror( errno ) );
+		else
+			line = world_msg_client( wld,
+					"Client closed connection." );
+
 		world_disconnect_client( wld );
+
+		line->flags = LINE_LOGONLY;
 		return;
 	}
 
@@ -893,8 +1006,9 @@ static void handle_client_fd( World *wld )
  * append to RX queue. If the connection died, close FD. */
 static void handle_server_fd( World *wld )
 {
-	int err, n, offset = wld->server_rxfull;
+	int n, offset = wld->server_rxfull;
 	char *buffer = wld->server_rxbuffer;
+	Line *line;
 
 	/* Read into buffer */
 	n = read( wld->server_fd, buffer + offset, NET_BBUFFER_LEN - offset );
@@ -903,20 +1017,19 @@ static void handle_server_fd( World *wld )
 	if( n == -1 && ( errno == EINTR || errno == EAGAIN ) )
 		return;
 
+	/* The connection died, notify the client */
 	if( n < 1 )
 	{
-		/* Ok, the connection dropped dead */
-		err = errno;
-		world_disconnect_server( wld );
-
-		/* Notify the client */
 		if( n < 0 )
-			world_msg_client( wld, "Connection to server "
-					"lost (%s).", strerror( err ) );
+			line = world_msg_client( wld, "Connection to server "
+					"lost (%s).", strerror( errno ) );
 		else
-			world_msg_client( wld,
+			line = world_msg_client( wld,
 					"The server closed the connection." );
 
+		world_disconnect_server( wld );
+
+		line->flags = LINE_CHECKPOINT;
 		wld->flags |= WLD_RECONNECT;
 		return;
 	}
@@ -936,7 +1049,8 @@ extern void world_flush_client_txbuf( World *wld )
 
 	flush_buffer( wld->client_fd, wld->client_txbuffer,
 			&wld->client_txfull, wld->client_txqueue, 
-			wld->inactive_lines, 1, NULL );
+			wld->inactive_lines, 1, wld->ace_prestr,
+			wld->ace_poststr, NULL );
 }
 
 
@@ -957,5 +1071,64 @@ extern void world_flush_server_txbuf( World *wld )
 
 	flush_buffer( wld->server_fd, wld->server_txbuffer,
 			&wld->server_txfull, wld->server_txqueue, NULL,
-			1, NULL );
+			1, NULL, NULL, NULL );
+}
+
+
+
+extern void world_auth_add_bucket( World *wld )
+{
+	wld->auth_tokenbucket += NET_AUTH_TOKENSPERSEC;
+	if( wld->auth_tokenbucket > NET_AUTH_BUCKETSIZE )
+		wld->auth_tokenbucket = NET_AUTH_BUCKETSIZE;
+}
+
+
+
+/* Add addr to the list of privileged addresses. Make room if needed.
+ * Ensure that the new addr is at the bottom, so the list is LRU-ordered. */
+static void privileged_add( World *wld, char *addr )
+{
+	/* If the address is already in the list, remove it.
+	 * This way, addr will be the last entry in the list, so the list
+	 * maintains a LRU ordering. */
+	privileged_del( wld, addr );
+
+	/* Make room if the list is full. */
+	if( wld->auth_privaddrs->count >= NET_MAX_PRIVADDRS )
+		line_destroy( linequeue_pop( wld->auth_privaddrs ) );
+
+	/* Add it. */
+	linequeue_append( wld->auth_privaddrs,
+			line_create( xstrdup( addr ), -1 ) );
+}
+
+
+
+/* Remove addr from the privileged addresses list, if present. */
+static void privileged_del( World *wld, char *addr )
+{
+	Line *line;
+
+	for( line = wld->auth_privaddrs->head; line; line = line->next )
+		if( !strcmp( addr, line->str ) )
+		{
+			linequeue_remove( wld->auth_privaddrs, line );
+			line_destroy( line );
+			return;
+		}
+}
+
+
+
+/* Check if addr is in the privileged address list. Returns 1 if so, else 0. */
+static int is_privileged( World *wld, char *addr )
+{
+	Line *line;
+
+	for( line = wld->auth_privaddrs->head; line; line = line->next )
+		if( !strcmp( addr, line->str ) )
+			return 1;
+
+	return 0;
 }
