@@ -1,7 +1,7 @@
 /*
  *
- *  mooproxy - a buffering proxy for moo-connections
- *  Copyright (C) 2001-2007 Marcel L. Moreaux <marcelm@luon.net>
+ *  mooproxy - a buffering proxy for MOO connections
+ *  Copyright (C) 2001-2011 Marcel L. Moreaux <marcelm@qvdr.net>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -25,14 +25,15 @@
 #define PANIC_MALLOC 2 
 #define PANIC_REALLOC 3
 #define PANIC_STRDUP 4
-#define PANIC_VASPRINTF 5
-#define PANIC_SELECT 6
-#define PANIC_ACCEPT 7
+#define PANIC_STRNDUP 5
+#define PANIC_VASPRINTF 6
+#define PANIC_SELECT 7
+#define PANIC_ACCEPT 8
 
 
 
 /* Signal handler. Basically calls panic(). */
-extern void panic_sighandler( int sig );
+extern void sighandler_panic( int sig );
 
 /* Panic. Try to write a helpful message to stderr, crash-file, and any
  * connected client. After that, terminate.
