@@ -1,7 +1,7 @@
 /*
  *
- *  mooproxy - a buffering proxy for moo-connections
- *  Copyright (C) 2001-2007 Marcel L. Moreaux <marcelm@luon.net>
+ *  mooproxy - a buffering proxy for MOO connections
+ *  Copyright (C) 2001-2011 Marcel L. Moreaux <marcelm@qvdr.net>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -55,8 +55,8 @@ MODIFIER  -> 'lines'
 
 
 
-typedef struct _Params Params;
-struct _Params
+typedef struct Params Params;
+struct Params
 {
 	/* The argument string we're parsing. */
 	char   *argstr;
@@ -424,9 +424,9 @@ static int parse_keyword_search( World *wld, Params *params )
 
 
 
-/* Parse a timespec. Returns true on error, false on succes.
+/* Parse a timespec. Returns true on error, false on success.
  * On error, params->error may or may not be set.
- * On succes, params->when or params->lines will be modified. */
+ * On success, params->when or params->lines will be modified. */
 static int parse_when( World *wld, Params *params, int lma )
 {
 	if( params->word[0] == '-' || params->word[0] == '+' )
@@ -437,9 +437,9 @@ static int parse_when( World *wld, Params *params, int lma )
 
 
 
-/* Parse a relative timespec. Returns true on error, false on succes.
+/* Parse a relative timespec. Returns true on error, false on success.
  * On error, params->error may or may not be set.
- * On succes, params->when or params->lines will be modified.
+ * On success, params->when or params->lines will be modified.
  * lma idicates whether or not the 'lines' modifier is allowed now. */
 static int parse_when_relative( World *wld, Params *params, int lma )
 {
@@ -540,9 +540,9 @@ static int parse_when_relative( World *wld, Params *params, int lma )
 
 
 
-/* Parse an absolute timespec. Returns true on error, false on succes.
+/* Parse an absolute timespec. Returns true on error, false on success.
  * On error, params->error may or may not be set.
- * On succes, params->when or params->lines will be modified. */
+ * On success, params->when or params->lines will be modified. */
 static int parse_when_absolute( World *wld, Params *params )
 {
 	struct tm *tm;
@@ -641,7 +641,7 @@ static int parse_when_absolute( World *wld, Params *params )
 
 
 /* Parse an absolute date, [YY/]MM/DD.
- * Return true on error, false on succes.
+ * Return true on error, false on success.
  * On success, params->when is modified. */
 static int parse_when_absdate( World *wld, Params *params )
 {
@@ -694,7 +694,7 @@ static int parse_when_absdate( World *wld, Params *params )
 
 
 /* Parse an absolute time, HH:MM[:SS].
- * Return true on error, false on succes.
+ * Return true on error, false on success.
  * On success, params->when is modified. */
 static int parse_when_abstime( World *wld, Params *params )
 {
