@@ -1,7 +1,7 @@
 /*
  *
- *  mooproxy - a buffering proxy for moo-connections
- *  Copyright (C) 2001-2007 Marcel L. Moreaux <marcelm@luon.net>
+ *  mooproxy - a buffering proxy for MOO connections
+ *  Copyright (C) 2001-2011 Marcel L. Moreaux <marcelm@qvdr.net>
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -22,6 +22,8 @@
 #include <time.h>
 #include <stdarg.h>
 #include <unistd.h>
+#include <stdio.h>
+#include <ctype.h>
 
 #include "global.h"
 #include "world.h"
@@ -32,6 +34,10 @@
 
 
 
+#define EASTEREGG_TRIGGER " vraagt aan je, \"Welke mooproxy versie draai je?\""
+
+
+
 static int worlds_count = 0;
 static World **worlds_list = NULL;
 
@@ -69,15 +75,17 @@ extern World *world_create( char *wldname )
 	world_bindresult_init( wld->bindresult );
 
 	/* Authentication related stuff */
-	wld->auth_md5hash = NULL;
+	wld->auth_hash = NULL;
 	wld->auth_literal = NULL;
+	wld->auth_tokenbucket = NET_AUTH_BUCKETSIZE;
 	wld->auth_connections = 0;
-	for( i = NET_MAXAUTHCONN - 1; i >= 0; i-- )
+	for( i = 0; i < NET_MAXAUTHCONN; i++ )
 	{
 		wld->auth_buf[i] = xmalloc( NET_MAXAUTHLEN );
 		wld->auth_address[i] = NULL;
 		wld->auth_fd[i] = -1;
 	}
+	wld->auth_privaddrs = linequeue_create();
 
 	/* Data related to the server connection */
 	wld->server_status = ST_DISCONNECTED;
@@ -98,9 +106,9 @@ extern World *world_create( char *wldname )
 	wld->server_rxqueue = linequeue_create();
 	wld->server_toqueue = linequeue_create();
 	wld->server_txqueue = linequeue_create();
-	wld->server_rxbuffer = xmalloc( NET_BBUFFER_LEN + 4 ); /* See (1) */
+	wld->server_rxbuffer = xmalloc( NET_BBUFFER_ALLOC ); /* See (1) */
 	wld->server_rxfull = 0;
-	wld->server_txbuffer = xmalloc( NET_BBUFFER_LEN + 4 ); /* See (1) */
+	wld->server_txbuffer = xmalloc( NET_BBUFFER_ALLOC ); /* See (1) */
 	wld->server_txfull = 0;
 
 	/* Data related to the client connection */
@@ -109,14 +117,17 @@ extern World *world_create( char *wldname )
 	wld->client_address = NULL;
 	wld->client_prev_address = NULL;
 	wld->client_last_connected = 0;
+	wld->client_login_failures = 0;
+	wld->client_last_failaddr = NULL;
+	wld->client_last_failtime = 0;
 	wld->client_last_notconnmsg = 0;
 
 	wld->client_rxqueue = linequeue_create();
 	wld->client_toqueue = linequeue_create();
 	wld->client_txqueue = linequeue_create();
-	wld->client_rxbuffer = xmalloc( NET_BBUFFER_LEN + 4 ); /* See (1) */
+	wld->client_rxbuffer = xmalloc( NET_BBUFFER_ALLOC ); /* See (1) */
 	wld->client_rxfull = 0;
-	wld->client_txbuffer = xmalloc( NET_BBUFFER_LEN + 4 ); /* See (1) */
+	wld->client_txbuffer = xmalloc( NET_BBUFFER_ALLOC ); /* See (1) */
 	wld->client_txfull = 0;
 
 	/* Miscellaneous */
@@ -125,6 +136,7 @@ extern World *world_create( char *wldname )
 	wld->history_lines = linequeue_create();
 	wld->dropped_inactive_lines = 0;
 	wld->dropped_buffered_lines = 0;
+	wld->easteregg_last = 0;
 
 	/* Timer stuff */
 	wld->timer_prev_sec = -1;
@@ -142,7 +154,7 @@ extern World *world_create( char *wldname )
 	wld->log_currenttime = 0;
 	wld->log_currenttimestr = NULL;
 	wld->log_fd = -1;
-	wld->log_buffer = xmalloc( NET_BBUFFER_LEN + 4 ); /* See (1) */
+	wld->log_buffer = xmalloc( NET_BBUFFER_ALLOC ); /* See (1) */
 	wld->log_bfull = 0;
 	wld->log_lasterror = NULL;
 	wld->log_lasterrtime = 0;
@@ -152,20 +164,26 @@ extern World *world_create( char *wldname )
 	wld->mcp_key = NULL;
 	wld->mcp_initmsg = NULL;
 
+	/* Ansi Client Emulation (ACE) */
+	wld->ace_enabled = 0;
+	wld->ace_prestr = NULL;
+	wld->ace_poststr = NULL;
+
 	/* Options */
-	wld->logging_enabled = DEFAULT_LOGENABLE;
 	wld->autologin = DEFAULT_AUTOLOGIN;
 	wld->autoreconnect = DEFAULT_AUTORECONNECT;
 	wld->commandstring = xstrdup( DEFAULT_CMDSTRING );
+	wld->strict_commands = DEFAULT_STRICTCMDS;
 	wld->infostring = xstrdup( DEFAULT_INFOSTRING );
 	wld->infostring_parsed = parse_ansi_tags( wld->infostring );
 	wld->newinfostring = xstrdup( DEFAULT_NEWINFOSTRING );
 	wld->newinfostring_parsed = parse_ansi_tags( wld->newinfostring );
-	wld->context_on_connect = DEFAULT_CONTEXTLINES;
-	wld->max_buffer_size = DEFAULT_MAXBUFFERSIZE;
-	wld->max_logbuffer_size = DEFAULT_MAXLOGBUFFERSIZE;
-	wld->strict_commands = DEFAULT_STRICTCMDS;
+	wld->context_lines = DEFAULT_CONTEXTLINES;
+	wld->buffer_size = DEFAULT_BUFFERSIZE;
+	wld->logbuffer_size = DEFAULT_LOGBUFFERSIZE;
+	wld->logging = DEFAULT_LOGGING;
 	wld->log_timestamps = DEFAULT_LOGTIMESTAMPS;
+	wld->easteregg_version = DEFAULT_EASTEREGGS;
 
 	/* Add to the list of worlds */
 	register_world( wld );
@@ -174,13 +192,17 @@ extern World *world_create( char *wldname )
 
 	/* (1) comment for allocations of several buffers:
 	 * 
-	 * The TX and log buffers must be a little larger than specified, so
-	 * that flush_buffer() can append a newline (\n or \r\n) to the line
-	 * written in the buffer, even if the buffer is completely filled with
-	 * lines.
+	 * We allocate using NET_BBUFFER_ALLOC rather than NET_BUFFER_LEN,
+	 * because the buffers need to be longer than their "pretend length",
+	 * so we can fit in some additional things even when the buffer is
+	 * completely full with lines.
 	 *
-	 * The RX buffers must be larger so that buffer_to_lines() can append a
-	 * sentinal for its linear search.
+	 * Some things that we need to fit in there:
+	 *
+	 *   - For the log buffer, a \n after the last line.
+	 *   - For the TX buffers, a \r\n and possible ansi sequences for
+	 *     client emulation
+	 *   - For the RX buffers, a sentinel for the linear search
 	 */
 }
 
@@ -202,20 +224,24 @@ extern void world_destroy( World *wld )
 	free( wld->dest_host );
 
 	/* Listening connection */
+	if( wld->listen_fds )
+		for( i = 0; wld->listen_fds[i] > -1; i++ )
+			close( wld->listen_fds[i] );
 	free( wld->listen_fds );
 	world_bindresult_free( wld->bindresult );
 	free( wld->bindresult );
 
 	/* Authentication related stuff */
-	free( wld->auth_md5hash );
+	free( wld->auth_hash );
 	free( wld->auth_literal );
-	for( i = NET_MAXAUTHCONN - 1; i >= 0; i-- )
+	for( i = 0; i < NET_MAXAUTHCONN; i++ )
 	{
 		free( wld->auth_buf[i] );
 		free( wld->auth_address[i] );
 		if( wld->auth_fd[i] > -1 )
 			close( wld->auth_fd[i] );
 	}
+	linequeue_destroy( wld->auth_privaddrs );
 
 	/* Data related to server connection */
 	if( wld->server_fd > -1 )
@@ -266,6 +292,10 @@ extern void world_destroy( World *wld )
 	free( wld->mcp_key );
 	free( wld->mcp_initmsg );
 
+	/* Ansi Client Emulation (ACE) */
+	free( wld->ace_prestr );
+	free( wld->ace_poststr );
+
 	/* Options */
 	free( wld->commandstring );
 	free( wld->infostring );
@@ -346,8 +376,8 @@ extern void world_bindresult_free( BindResult *bindresult )
 	for( i = 0; i < bindresult->af_count; i++ )
 		free( bindresult->af_msg[i] );
 	free( bindresult->af_msg );
-	if( bindresult->listen_fds != NULL )
-		for( i = 0; bindresult->listen_fds[i] != -1; i++ )
+	if( bindresult->listen_fds )
+		for( i = 0; bindresult->listen_fds[i] > -1; i++ )
 			close( bindresult->listen_fds[i] );
 	free( bindresult->listen_fds );
 	free( bindresult->conclusion );
@@ -396,7 +426,7 @@ extern Line *world_newmsg_client( World *wld, const char *fmt, ... )
 
 
 
-/* Send prefix + str + MESSAGE_TERMINATOR to the client.
+/* Send prefix + str + <ansi reset> to the client.
  * str is consumed, prefix is not.
  * If prefix is NULL or "", use wld->infostring_parsed.
  * Returns the line object that has been enqueued to the client. */
@@ -408,12 +438,11 @@ static Line *message_client( World *wld, char *prefix, char *str )
 	if( prefix == NULL || *prefix == '\0' )
 		prefix = wld->infostring_parsed;
 
-	/* Construct the message */
-	msg = xmalloc( strlen( str ) + strlen( prefix ) +
-			sizeof( MESSAGE_TERMINATOR ) );
+	/* Construct the message. The 5 is for the ansi reset and \0. */
+	msg = xmalloc( strlen( str ) + strlen( prefix ) + 5 );
 	strcpy( msg, prefix );
 	strcat( msg, str );
-	strcat( msg, MESSAGE_TERMINATOR );
+	strcat( msg, "\x1B[0m" );
 	free( str );
 
 	/* And append it to the client queue */
@@ -436,7 +465,7 @@ extern void world_trim_dynamic_queues( World *wld )
 	/* Trim the normal buffers. The data is distributed over
 	 * buffered_lines, inactive_lines and history_lines. We will trim
 	 * them in reverse order until everything is small enough. */
-	limit = wld->max_buffer_size * 1024;
+	limit = wld->buffer_size * 1024;
 
 	/* First, the history lines. Remove oldest lines. */
 	while( *bll + *ill + *hll > limit && wld->history_lines->head != NULL )
@@ -461,7 +490,7 @@ extern void world_trim_dynamic_queues( World *wld )
 	/* Trim the logbuffers. The data is distributed over log_current and
 	 * log_queue. We will trim them in reverse order until everything is
 	 * small enough. */
-	limit = wld->max_logbuffer_size * 1024;
+	limit = wld->logbuffer_size * 1024;
 
 	/* First log_queue. These are very important, so we count the number
 	 * of dropped lines. Remove newest lines. */
@@ -492,10 +521,10 @@ extern void world_recall_and_pass( World *wld )
 	Line *line, *recalled;
 	int nocontext = 0, noposnew = 0, nocernew = 0;
 
-	if( wld->context_on_connect > 0 )
+	if( wld->context_lines > 0 )
 	{
 		/* Recall history lines. */
-		queue = world_recall_history( wld, wld->context_on_connect );
+		queue = world_recall_history( wld, wld->context_lines );
 
 		/* Inform the client of the context lines */
 		if( queue->count > 0 )
@@ -523,7 +552,7 @@ extern void world_recall_and_pass( World *wld )
 			"(%.1f%% of buffer).",
 			( nocontext == 0 ) ? "" : "No context lines. ",
 			il->count, ( il->count == 1 ) ? "" : "s",
-			il->size / 10.24 / wld->max_buffer_size );
+			il->size / 10.24 / wld->buffer_size );
 
 		/* If nocontext was 1, we reported on it now. Clear it. */
 		nocontext = 0;
@@ -552,7 +581,7 @@ extern void world_recall_and_pass( World *wld )
 			( nocontext == 0 ) ? "" : "No context lines. ",
 			( noposnew == 0 ) ? "" : "No possibly new lines. ",
 			bl->count, ( bl->count == 1 ) ? "" : "s",
-			bl->size / 10.24 / wld->max_buffer_size );
+			bl->size / 10.24 / wld->buffer_size );
 
 		/* If nocontext/noposnew were 1, we reported. Clear them. */
 		nocontext = 0;
@@ -749,7 +778,7 @@ extern void world_rebind_port( World *wld )
 	/* Update the listenport. */
 	wld->listenport = wld->requestedlistenport;
 
-	/* Announce succes. */
+	/* Announce success. */
 	world_msg_client( wld, "The option listenport has been changed "
 			"to %li.", wld->listenport );
 }
@@ -790,10 +819,10 @@ extern void world_schedule_reconnect( World *wld )
 	free( whenstr );
 
 	/* Increase the delay. */
-	wld->reconnect_delay *= 2;
-	wld->reconnect_delay += 4000 + rand() / (RAND_MAX / 3000);
-	if( wld->reconnect_delay > 3600000 )
-		wld->reconnect_delay = 3600000;
+	wld->reconnect_delay *= AUTORECONNECT_BACKOFF_FACTOR;
+	wld->reconnect_delay += 3000 + rand() / (RAND_MAX / 4000);
+	if( wld->reconnect_delay > AUTORECONNECT_MAX_DELAY * 1000 )
+		wld->reconnect_delay = AUTORECONNECT_MAX_DELAY * 1000;
 
 	/* If we have to reconnect immediately, actually do it immediately. */
 	if( delay == 0 )
@@ -812,10 +841,9 @@ extern void world_do_reconnect( World *wld )
 	wld->flags |= WLD_SERVERRESOLVE;
 
 	/* And announce what we did. */
-	linequeue_append( wld->client_toqueue,
-			line_create( xstrdup( "" ), -1 ) );
+	world_msg_client( wld, "" );
 	world_msg_client( wld, "Reconnecting now (at %s).", 
-			time_string( current_time(), FULL_TIME ) );
+			time_fullstr( current_time() ) );
 }
 
 
@@ -827,3 +855,191 @@ extern void world_decrease_reconnect_delay( World *wld )
 	if( wld->reconnect_delay < 0 )
 		wld->reconnect_delay = 0;
 }
+
+
+
+extern int world_enable_ace( World *wld )
+{
+	Linequeue *queue;
+	char *tmp, *status;
+	int cols = wld->ace_cols, rows = wld->ace_rows;
+
+	/* Create the "statusbar". */
+	status = xmalloc( cols + strlen( wld->name ) + 20 );
+	sprintf( status, "---- mooproxy - %s ", wld->name );
+	memset( status + strlen( wld->name ) + 17, '-', cols );
+	status[cols] = '\0';
+
+	/* Construct the string to send to the client:
+	 *   - reset terminal
+	 *   - move cursor to the statusbar location
+	 *   - print statusbar
+	 *   - set scroll area to "input window"
+	 *   - move cursor to the very last line.
+	 */
+	xasprintf( &tmp, "\x1B" "c\x1B[%i;1H%s\x1B[%i;%ir\x1B[%i;1H",
+			rows - 3, status, rows - 2, rows, rows );
+	free( status );
+
+	/* We fail if the string doesn't fit into the buffer. */
+	if( wld->client_txfull + strlen( tmp ) > NET_BBUFFER_ALLOC )
+	{
+		free( tmp );
+		return 0;
+	}
+
+	/* Append the string to the buffer. */
+	strcpy( wld->client_txbuffer + wld->client_txfull, tmp );
+	wld->client_txfull += strlen( tmp );
+	free( tmp );
+
+	/* Construct the string that will be prepended to each line:
+	 *   - save cursor position
+	 *   - set scroll area to "output window"
+	 *   - set cursor to the last line of the "output window"
+	 */
+	free( wld->ace_prestr );
+	xasprintf( &wld->ace_prestr, "\x1B[s\x1B[1;%ir\x1B[%i;1H",
+			rows - 4, rows - 4 );
+
+	/* Construct the string that will be appended to each line:
+	 *   - set scroll area to "input window"
+	 *   - restore cursor position
+	 *   - reset attribute mode
+	 */
+	free( wld->ace_poststr );
+	xasprintf( &wld->ace_poststr, "\x1B[%i;%ir\x1B[u\x1B[0m",
+			rows - 2, rows );
+
+	/* Of course, ACE is enabled now. */
+	wld->ace_enabled = 1;
+
+	/* We want to include the inactive lines in our recall as well, so
+	 * we move them to the history right now. */
+	world_inactive_to_history( wld );
+
+	/* Recall enough lines to fill the "output window". */
+	queue = world_recall_history( wld, rows - 4 );
+	linequeue_merge( wld->client_toqueue, queue );
+	linequeue_destroy( queue );
+
+	return 1;
+}
+
+
+
+extern void world_disable_ace( World *wld )
+{
+	/* Send "reset terminal" to the client, so the clients terminal is
+	 * not left in a messed up state.
+	 * ACE deactivation should always succeed, so if the ansi sequence
+	 * doesn't fit in the buffer, we'll still do all the other stuff. */
+	if( wld->client_txfull + 2 < NET_BBUFFER_ALLOC )
+	{
+		wld->client_txbuffer[wld->client_txfull++] = '\x1B';
+		wld->client_txbuffer[wld->client_txfull++] = 'c';
+	}
+
+	free( wld->ace_prestr );
+	wld->ace_prestr = NULL;
+
+	free( wld->ace_poststr );
+	wld->ace_poststr = NULL;
+
+	wld->ace_enabled = 0;
+}
+
+
+
+/* A little easter egg. If someone asks us (in Dutch, in a specific format)
+ * what Mooproxy version we're running, we respond. */
+extern void world_easteregg_server( World *wld, Line *line )
+{
+	const static char *verbs[] =
+		{"an", "bl", "fl", "gie", "gl", "gn", "gr", "gri",
+		"grm", "ju", "ko", "mo", "mu", "pe", "sn", "ve", "zu", "zwi"};
+	const char *verb;
+	char *name, *response, *p;
+
+	/* Only once per minute, please. */
+	if( wld->easteregg_last > current_time() - 60 )
+		return;
+
+	/* If the line is too short, it can't be the trigger. */
+	if( line->len < sizeof( EASTEREGG_TRIGGER ) - 1 + 1 )
+		return;
+
+	/* Trigger? */
+	if( strcmp( line->str + line->len - sizeof( EASTEREGG_TRIGGER ) + 1,
+			EASTEREGG_TRIGGER ) )
+		return;
+
+	/* Isolate the name of the player triggering the easter egg. */
+	name = xmalloc( line->len + 1 );
+	strcpy_noansi( name, line->str );
+	for( p = name; *p; p++ )
+		if( isspace( *p ) )
+			break;
+	*p = '\0';
+
+	if( name[0] == '[' || name[0] == '<' || name[0] == '{' )
+	{
+		/* Looks like a channel, let's ignore it. */
+		free( name );
+		return;
+	}
+
+	/* Pick a random verb. */
+	verb = verbs[rand() % ( sizeof( verbs ) / sizeof( char * ) )];
+
+	if( name[0] == '>' && name[1] == '>' )
+		/* Response to a page. */
+		xasprintf( &response, "> %s Ik draai mooproxy %s",
+				verb, VERSIONSTR );
+	else
+		/* Response to a local player. */
+		xasprintf( &response, "%s -%s Ik draai mooproxy %s",
+				verb, name, VERSIONSTR );
+
+	/* Send the message and clean up. */
+	linequeue_append( wld->server_toqueue, line_create( response, -1 ) );
+	wld->easteregg_last = current_time();
+	free( name );
+}
+
+
+
+extern int world_start_shutdown( World *wld, int force, int fromsig )
+{
+	Line *line;
+
+	/* Unlogged data? Refuse if not forced. */
+	if( !force && wld->log_queue->size + wld->log_current->size +
+			wld->log_bfull > 0 )
+	{
+		long unlogged_lines = 1 + wld->log_queue->count +
+				wld->log_current->count + wld->log_bfull / 80;
+		long unlogged_kib = ( wld->log_bfull + wld->log_queue->size +
+				wld->log_current->size + 512 ) / 1024;
+		world_msg_client( wld, "There are approximately %li lines "
+				"(%liKiB) not yet logged to disk. ",
+				unlogged_lines, unlogged_kib );
+		world_msg_client( wld, "Refusing to shut down. "
+				"Use %s to override.",
+				fromsig ? "SIGQUIT" : "shutdown -f" );
+		return 0;
+	}
+
+	/* We try to leave their terminal in a nice state. */
+	if( wld->ace_enabled )
+		world_disable_ace( wld );
+
+	/* We're good, proceed. */
+	line = world_msg_client( wld, "Shutting down%s (reason: %s).",
+			force ? " forcibly" : "",
+			fromsig ? "signal" : "client request" );
+	line->flags = LINE_CHECKPOINT;
+	wld->flags |= WLD_SHUTDOWN;
+
+	return 1;
+}
